/*
  Copyright 2019 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.auth

import com.kakao.sdk.auth.model.AccessTokenResponse
import com.kakao.sdk.auth.model.AgtResponse
import com.kakao.sdk.auth.model.OAuthToken
import com.kakao.sdk.auth.network.kauth
import com.kakao.sdk.common.network.ApiFactory
import com.kakao.sdk.common.util.KakaoJson
import com.kakao.sdk.common.KakaoSdk
import com.kakao.sdk.common.model.*
import retrofit2.Call
import retrofit2.Callback
import retrofit2.HttpException
import retrofit2.Response

/**
 * 카카오 OAuth 서버에서 제공하는 API 를 사용하기 위한 클라이언트.
 */
class AuthApiClient(
    private val authApi: AuthApi = ApiFactory.kauth.create(AuthApi::class.java),
    private val tokenManagerProvider: TokenManagerProvider = TokenManagerProvider.instance,
    private val applicationInfo: ApplicationInfo = KakaoSdk.applicationContextInfo,
    private val contextInfo: ContextInfo = KakaoSdk.applicationContextInfo
) {

    /**
     * [AuthCodeClient] 를 이용하여 발급 받은 인증 코드를 사용하여 액세스 토큰을 발급.
     *
     * 발급된 토큰은 [TokenManagerProvider]에 지정된 토큰 저장소에 자동으로 저장됨.
     *
     * @param authCode 인증 코드.
     * @param callback [OAuthToken] 반환.
     */
    @JvmOverloads
    fun issueAccessToken(
        authCode: String,
        callback: (token: OAuthToken?, error: Throwable?) -> Unit
    ) {
        val clientId = applicationInfo.appKey
        val redirectUri = "kakao$clientId://oauth"
        val androidKeyHash = contextInfo.signingKeyHash
        authApi.issueAccessToken(
            clientId = clientId,
            redirectUri = redirectUri,
            androidKeyHash = androidKeyHash,
            authCode = authCode
        ).enqueue(object: Callback<AccessTokenResponse> {
            override fun onFailure(call: Call<AccessTokenResponse>, t: Throwable) {
                callback(null, t)
            }

            override fun onResponse(
                call: Call<AccessTokenResponse>,
                response: Response<AccessTokenResponse>
            ) {
                if (response.isSuccessful) {
                    response.body()?.let {
                        val token = OAuthToken.fromResponse(it)
                        tokenManagerProvider.manager.setToken(token)
                        callback(token, null)
                        return
                    }
                    callback(null, ClientError(ClientErrorCause.Unknown, "No body"))
                } else {
                    callback(null, translateError(HttpException(response)))
                }
            }
        })
    }

    /**
     * 리프레시 토큰으로 액세스 토큰을 갱신.
     *
     * 갱신된 토큰은 [TokenManagerProvider]에 지정된 토큰 저장소에 자동으로 저장됨.
     *
     * @param oldToken 이전에 발급 받았던 토큰. 기본 값은 [TokenManagerProvider]에 지정된 저장소의 토큰.
     * @param callback [OAuthToken] 반환.
     *
     * @throws ClientErrorCause.TokenNotFound
     */
    @JvmOverloads
    fun refreshAccessToken(
        oldToken: OAuthToken = tokenManagerProvider.manager.getToken() ?: throw ClientError(ClientErrorCause.TokenNotFound),
        callback: (token: OAuthToken?, error: Throwable?) -> Unit) {
        val clientId = applicationInfo.appKey
        val androidKeyHash = contextInfo.signingKeyHash
        authApi.issueAccessToken(
            clientId = clientId,
            redirectUri = null,
            androidKeyHash = androidKeyHash,
            refreshToken = oldToken.refreshToken,
            grantType = Constants.REFRESH_TOKEN
        ).enqueue(object : Callback<AccessTokenResponse> {
            override fun onFailure(call: Call<AccessTokenResponse>, t: Throwable) {
                callback(null, t)
            }

            override fun onResponse(
                call: Call<AccessTokenResponse>,
                response: Response<AccessTokenResponse>
            ) {
                if (response.isSuccessful) {
                    response.body()?.let {
                        val token = OAuthToken.fromResponse(it, oldToken)
                        tokenManagerProvider.manager.setToken(token)
                        callback(token, null)
                        return
                    }
                    callback(null, ClientError(ClientErrorCause.Unknown, "No body"))
                } else {
                    callback(null, translateError(HttpException(response)))
                }
            }

        })
    }

    /**
     * @suppress
     */
    fun agt(
        callback: (response: AgtResponse?, error: Throwable?) -> Unit
    ) {
        val clientId = applicationInfo.appKey
        tokenManagerProvider.manager.getToken()?.accessToken?.let {
            authApi.agt(clientId, it).enqueue(object: Callback<AgtResponse> {
                override fun onFailure(call: Call<AgtResponse>, t: Throwable) {
                    callback(null, t)
                }

                override fun onResponse(call: Call<AgtResponse>, response: Response<AgtResponse>) {
                    response.body()?.let {
                        callback(it, null)
                        return
                    }
                    callback(null, translateError(HttpException(response)))
                }
            })
            return
        }
        callback(null, ClientError(ClientErrorCause.TokenNotFound, "Access token not found."))
    }

    fun refreshAccessToken(
        oldToken: OAuthToken = tokenManagerProvider.manager.getToken() ?: throw ClientError(ClientErrorCause.TokenNotFound)
    ): OAuthToken {
        val clientId = applicationInfo.appKey
        val androidKeyHash = contextInfo.signingKeyHash
        val response =
            authApi.issueAccessToken(
                clientId = clientId,
                redirectUri = null,
                androidKeyHash = androidKeyHash,
                refreshToken = oldToken.refreshToken,
                grantType = Constants.REFRESH_TOKEN
            ).execute()

        val token = response.body()?.let {
            OAuthToken.fromResponse(it, oldToken)
        } ?: throw translateError(HttpException(response))

        tokenManagerProvider.manager.setToken(token)
        return token
    }

    companion object {
        /**
         * @suppress
         */
        fun translateError(t: Throwable): Throwable {
            try {
                if (t is HttpException) {
                    val errorString = t.response()?.errorBody()?.string()
                    val response =
                        KakaoJson.fromJson<AuthErrorResponse>(errorString!!, AuthErrorResponse::class.java)
                    val cause = KakaoJson.fromJson(response.error, AuthErrorCause::class.java)
                        ?: AuthErrorCause.Unknown
                    return AuthError(t.code(), cause, response)
                }
                return t
            } catch (unexpected: Throwable) {
                return unexpected
            }
        }

        /**
         * 간편한 API 호출을 위해 기본 제공되는 singleton 객체
         */
        @JvmStatic
        val instance by lazy { AuthApiClient() }
    }
}