/*
  Copyright 2020 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.auth.network

import com.kakao.sdk.auth.LoginClient
import com.kakao.sdk.auth.model.OAuthToken
import com.kakao.sdk.auth.network.AccessTokenInterceptor.Companion.requestWithAuthorization
import com.kakao.sdk.common.util.KakaoJson
import com.kakao.sdk.common.KakaoSdk
import com.kakao.sdk.common.model.ApiErrorCause
import com.kakao.sdk.common.model.ApiErrorResponse
import com.kakao.sdk.common.model.ApplicationContextInfo
import okhttp3.Interceptor
import okhttp3.Response
import okhttp3.ResponseBody
import java.util.concurrent.CountDownLatch

/**
 * @suppress
 *
 * -402 에러 시 자동 추가 동의
 */
class RequiredScopesInterceptor(
    private val contextInfo: ApplicationContextInfo = KakaoSdk.applicationContextInfo
) : Interceptor {

    override fun intercept(chain: Interceptor.Chain): Response {
        val originalResponse = chain.proceed(chain.request())

        if (originalResponse.isSuccessful) {
            return originalResponse
        }

        val body = originalResponse.body()
        val bodyString = body?.string()

        // 에러 응답 파싱을 위해 body를 한번 소비했기 때문에 새로 만들어주지 않으면 다른 인터셉터에서 body 접근 시 크래시 난다.
        val response = originalResponse.newBuilder()
            .body(ResponseBody.create(body?.contentType(), bodyString))
            .build()

        val apiErrorResponse = bodyString?.let {
            KakaoJson.fromJson<ApiErrorResponse>(it, ApiErrorResponse::class.java)
        }
        val apiErrorCause = apiErrorResponse?.let {
            KakaoJson.fromJson<ApiErrorCause>(it.code.toString(), ApiErrorCause::class.java)
        }
        val requiredScopes = apiErrorResponse?.requiredScopes

        if (apiErrorCause == ApiErrorCause.InvalidScope && requiredScopes != null) {
            var token: OAuthToken? = null
            var error: Throwable? = null

            val latch = CountDownLatch(1)
            LoginClient.instance.loginWithNewScopes(contextInfo.applicationContext, requiredScopes) { t, e -> token = t; error = e; latch.countDown() }
            latch.await()

            return token?.accessToken?.let {
                chain.proceed(
                    requestWithAuthorization(
                        response.request(),
                        it
                    )
                )
            } ?: throw java.io.IOException("Kakao SDK Error", error!!)
        }

        return response
    }
}